@@ -7,11 +7,15 @@ import shlex
 import subprocess
 import sys
 import urllib.error
+from dataclasses import asdict
 from datetime import datetime, timezone
+from functools import lru_cache
 from textwrap import dedent
-from typing import get_args
+from typing import Any, get_args
 from urllib.parse import urlparse
 
+from ramalama.benchmarks.manager import BenchmarksManager
+
 # if autocomplete doesn't exist, just do nothing, don't break
 try:
     import argcomplete
@@ -20,32 +24,34 @@ try:
 except Exception:
     suppressCompleter = None
 
-
-import ramalama.chat as chat
 from ramalama import engine
-from ramalama.chat import default_prefix
+from ramalama.arg_types import DefaultArgsType
+from ramalama.benchmarks.utilities import print_bench_results
 from ramalama.cli_arg_normalization import normalize_pull_arg
 from ramalama.command.factory import assemble_command
-from ramalama.common import accel_image, get_accel, perror
+from ramalama.common import accel_image, exec_cmd, get_accel, perror
 from ramalama.config import (
-    COLOR_OPTIONS,
-    CONFIG,
     GGUF_QUANTIZATION_MODES,
     SUPPORTED_ENGINES,
     SUPPORTED_RUNTIMES,
     coerce_to_bool,
+    get_config,
     get_inference_schema_files,
     get_inference_spec_files,
     load_file_config,
 )
+from ramalama.config_types import COLOR_OPTIONS
 from ramalama.endian import EndianMismatchError
 from ramalama.log_levels import LogLevel
 from ramalama.logger import configure_logger, logger
 from ramalama.model_inspect.error import ParseError
 from ramalama.model_store.global_store import GlobalModelStore
-from ramalama.rag import INPUT_DIR, Rag, RagTransport, rag_image
+from ramalama.path_utils import file_uri_to_path
+from ramalama.prompt_utils import default_prefix
+from ramalama.rag import RagTransport, rag_image
 from ramalama.shortnames import Shortnames
 from ramalama.stack import Stack
+from ramalama.transports.api import APITransport
 from ramalama.transports.base import (
     MODEL_TYPES,
     NoGGUFModelFileFound,
@@ -57,13 +63,30 @@ from ramalama.transports.base import (
 from ramalama.transports.transport_factory import New, TransportFactory
 from ramalama.version import print_version, version
 
-shortnames = Shortnames()
-
 GENERATE_OPTIONS = ["quadlet", "kube", "quadlet/kube", "compose"]
 LIST_SORT_FIELD_OPTIONS = ["size", "modified", "name"]
 LIST_SORT_ORDER_OPTIONS = ["desc", "asc"]
 
 
+@lru_cache(maxsize=1)
+def default_image() -> str:
+    return accel_image(get_config())
+
+
+@lru_cache(maxsize=1)
+def default_rag_image() -> str:
+    return rag_image(get_config())
+
+
+@lru_cache(maxsize=1)
+def get_shortnames():
+    return Shortnames()
+
+
+def assemble_command_lazy(cli_args: argparse.Namespace) -> list[str]:
+    return assemble_command(cli_args)
+
+
 class ParsedGenerateInput:
     def __init__(self, gen_type: str, output_dir: str):
         self.gen_type = gen_type
@@ -122,11 +145,13 @@ def local_models(prefix, parsed_args, **kwargs):
 
 def local_containers(prefix, parsed_args, **kwargs):
     parsed_args.format = '{{.Names}}'
+
     return engine.containers(parsed_args)
 
 
 def local_images(prefix, parsed_args, **kwargs):
     parsed_args.format = "{{.Repository}}:{{.Tag}}"
+
     return engine.images(parsed_args)
 
 
@@ -134,7 +159,7 @@ def available_metadata(prefix, parsed_args, **kwargs):
     if parsed_args.MODEL:
         # shotnames resolution has not been applied for auto-completion
         # Therefore it needs to be done explicitly here in order to support it
-        resolved_model = shortnames.resolve(parsed_args.MODEL)
+        resolved_model = get_shortnames().resolve(parsed_args.MODEL)
 
         metadata = New(resolved_model, parsed_args).inspect_metadata()
         return [field for field in metadata.keys() if field.startswith(parsed_args.get)]
@@ -151,39 +176,45 @@ class ArgumentParserWithDefaults(argparse.ArgumentParser):
                 kwargs['help'] += f' (default: {default})'
         action = super().add_argument(*args, **kwargs)
         if completer is not None:
-            action.completer = completer
+            action.completer = completer  # type: ignore[attr-defined]
         return action
 
 
+def get_initial_parser():
+    description = get_description()
+    parser = create_argument_parser(description, add_help=False)
+    return parser
+
+
 def get_parser():
     description = get_description()
-    parser = create_argument_parser(description)
+    parser = create_argument_parser(description, add_help=True)
     configure_subcommands(parser)
     return parser
 
 
 def init_cli():
     """Initialize the RamaLama CLI and parse command line arguments."""
-    # Need to know if we're running with --dryrun or --generate before adding the subcommands,
-    # otherwise calls to accel_image() when setting option defaults will cause unnecessary image pulls.
-    if any(arg in ("--dryrun", "--dry-run", "--generate") or arg.startswith("--generate=") for arg in sys.argv[1:]):
-        CONFIG.dryrun = True
-    parser = get_parser()
-    args = parse_arguments(parser)
-    post_parse_setup(args)
-    return parser, args
+    return parse_args_from_cmd(sys.argv[1:])
 
 
-def parse_args_from_cmd(cmd: list[str]) -> argparse.Namespace:
+def parse_args_from_cmd(cmd: list[str]) -> tuple[argparse.ArgumentParser, argparse.Namespace]:
     """Parse arguments based on a command string"""
-    # Need to know if we're running with --dryrun or --generate before adding the subcommands,
-    # otherwise calls to accel_image() when setting option defaults will cause unnecessary image pulls.
+    config = get_config()
     if any(arg in ("--dryrun", "--dry-run", "--generate") or arg.startswith("--generate=") for arg in sys.argv[1:]):
-        CONFIG.dryrun = True
+        config.dryrun = True
+    # Phase 1: Parse the initial arguments to set CONFIG.runtime etc... as this can affect the subcommands
+    initial_parser = get_initial_parser()
+    initial_args, _ = initial_parser.parse_known_args(cmd)
+    for arg in initial_args.__dict__.keys():
+        if hasattr(config, arg):
+            setattr(config, arg, getattr(initial_args, arg))
+
+    # Phase 2: Re-parse the arguments with the subcommands enabled
     parser = get_parser()
     args = parser.parse_args(cmd)
     post_parse_setup(args)
-    return args
+    return parser, args
 
 
 def get_description():
@@ -212,12 +243,13 @@ def abspath(astring) -> str:
     return os.path.abspath(astring)
 
 
-def create_argument_parser(description: str):
+def create_argument_parser(description: str, add_help: bool = True):
     """Create and configure the argument parser for the CLI."""
     parser = ArgumentParserWithDefaults(
         prog="ramalama",
         description=description,
         formatter_class=argparse.RawTextHelpFormatter,
+        add_help=add_help,
     )
     configure_arguments(parser)
     return parser
@@ -225,11 +257,12 @@ def create_argument_parser(description: str):
 
 def configure_arguments(parser):
     """Configure the command-line arguments for the parser."""
+    config = get_config()
     verbosity_group = parser.add_mutually_exclusive_group()
     parser.add_argument(
         "--container",
         dest="container",
-        default=CONFIG.container,
+        default=config.container,
         action="store_true",
         help=argparse.SUPPRESS,
     )
@@ -248,7 +281,7 @@ def configure_arguments(parser):
     parser.add_argument(
         "--engine",
         dest="engine",
-        default=CONFIG.engine,
+        default=config.engine,
         choices=get_args(SUPPORTED_ENGINES),
         help="""run RamaLama using the specified container engine.
 The RAMALAMA_CONTAINER_ENGINE environment variable modifies default behaviour.""",
@@ -256,21 +289,27 @@ The RAMALAMA_CONTAINER_ENGINE environment variable modifies default behaviour.""
     parser.add_argument(
         "--nocontainer",
         dest="container",
-        default=not CONFIG.container,
+        default=not config.container,
         action="store_false",
         help="""do not run RamaLama in the default container.
 The RAMALAMA_IN_CONTAINER environment variable modifies default behaviour.""",
     )
-    verbosity_group.add_argument("--quiet", "-q", dest="quiet", action="store_true", help="reduce output.")
+    verbosity_group.add_argument(
+        "--quiet",
+        "-q",
+        dest="quiet",
+        action="store_true",
+        help="Reduce output verbosity (silences warnings, simplifies list output)",
+    )
     parser.add_argument(
         "--runtime",
-        default=CONFIG.runtime,
+        default=config.runtime,
         choices=get_args(SUPPORTED_RUNTIMES),
         help="specify the runtime to use; valid options are 'llama.cpp', 'vllm', and 'mlx'",
     )
     parser.add_argument(
         "--store",
-        default=CONFIG.store,
+        default=config.store,
         type=abspath,
         help="store AI Models in the specified directory",
     )
@@ -285,6 +324,7 @@ def configure_subcommands(parser):
     subparsers = parser.add_subparsers(dest="subcommand")
     subparsers.required = False
     bench_parser(subparsers)
+    benchmarks_parser(subparsers)
     chat_parser(subparsers)
     containers_parser(subparsers)
     convert_parser(subparsers)
@@ -306,15 +346,12 @@ def configure_subcommands(parser):
     daemon_parser(subparsers)
 
 
-def parse_arguments(parser):
-    """Parse command line arguments."""
-    return parser.parse_args()
-
-
 def post_parse_setup(args):
     """Perform additional setup after parsing arguments."""
+    if getattr(args, "subcommand", None) == "benchmark":
+        args.subcommand = "bench"
 
-    def map_https_to_transport(input: str) -> str | None:
+    def map_https_to_transport(input: str) -> str:
         if input.startswith("https://") or input.startswith("http://"):
             url = urlparse(input)
             # detect if the whole repo is defined or a specific file
@@ -330,6 +367,7 @@ def post_parse_setup(args):
     # First, map https:// inputs to ollama or huggingface based on url domain
     # Then resolve the input based on the available shortname list
     if getattr(args, "MODEL", None):
+        shortnames = get_shortnames()
         if isinstance(args.MODEL, str):
             args.INITIAL_MODEL = args.MODEL
             args.MODEL = map_https_to_transport(args.MODEL)
@@ -372,7 +410,13 @@ def post_parse_setup(args):
     if hasattr(args, 'pull'):
         args.pull = normalize_pull_arg(args.pull, getattr(args, 'engine', None))
 
-    log_level = LogLevel.DEBUG if args.debug else (CONFIG.log_level or LogLevel.WARNING)
+    # Determine log level: debug > quiet > config > default
+    if args.debug:
+        log_level = LogLevel.DEBUG
+    elif getattr(args, 'quiet', False):
+        log_level = LogLevel.ERROR
+    else:
+        log_level = get_config().log_level or LogLevel.WARNING
     configure_logger(log_level)
 
 
@@ -398,7 +442,7 @@ def login_parser(subparsers):
 
 
 def normalize_registry(registry):
-    registry = registry or CONFIG.transport
+    registry = registry or get_config().transport
 
     if not registry or registry == "" or registry.startswith("oci://"):
         return "oci://"
@@ -428,6 +472,7 @@ def logout_parser(subparsers):
 
 def logout_cli(args):
     registry = normalize_registry(args.REGISTRY)
+
     model = New(registry, args)
     return model.logout(args)
 
@@ -463,12 +508,13 @@ def human_duration(d):
 
 
 def bench_cli(args):
+
     model = New(args.MODEL, args)
     model.ensure_model_exists(args)
-    model.bench(args, assemble_command(args))
+    model.bench(args, assemble_command_lazy(args))
 
 
-def add_network_argument(parser, dflt="none"):
+def add_network_argument(parser, dflt: str | None = "none"):
     # Disable network access by default, and give the option to pass any supported network mode into
     # podman if needed:
     # https://docs.podman.io/en/latest/markdown/podman-run.1.html#network-mode-net
@@ -493,10 +539,68 @@ def add_network_argument(parser, dflt="none"):
 def bench_parser(subparsers):
     parser = subparsers.add_parser("bench", aliases=["benchmark"], help="benchmark specified AI Model")
     runtime_options(parser, "bench")
-    parser.add_argument("MODEL", completer=local_models)  # positional argument
+    parser.add_argument("MODEL", completer=local_models)
+    parser.add_argument(
+        "--format",
+        choices=["table", "json"],
+        default="table",
+        help="output format (table or json)",
+    )
     parser.set_defaults(func=bench_cli)
 
 
+def benchmarks_parser(subparsers):
+    config = get_config()
+    storage_folder = config.benchmarks.storage_folder
+    epilog = f"Storage folder: {storage_folder}" if storage_folder else "Storage folder: not configured"
+    parser = subparsers.add_parser(
+        "benchmarks",
+        help="manage and view benchmark results",
+        epilog=epilog,
+    )
+    parser.set_defaults(func=lambda _: parser.print_help())
+
+    benchmarks_subparsers = parser.add_subparsers(dest="benchmarks_command", metavar="[command]")
+
+    list_parser = benchmarks_subparsers.add_parser("list", help="list benchmark results")
+    list_parser.add_argument(
+        "--limit",
+        type=int,
+        default=None,
+        help="limit number of results to display",
+    )
+    list_parser.add_argument(
+        "--offset",
+        type=int,
+        default=0,
+        help="offset for pagination",
+    )
+    list_parser.add_argument(
+        "--format",
+        choices=["table", "json"],
+        default="table",
+        help="output format (table or json)",
+    )
+    list_parser.set_defaults(func=benchmarks_list_cli)
+
+
+def benchmarks_list_cli(args):
+    """Display a list of benchmark results from storage."""
+    config = get_config()
+    bench_manager = BenchmarksManager(config.benchmarks.storage_folder)
+    results = bench_manager.list()
+
+    if not results:
+        print("No benchmark results found")
+        return
+
+    if args.format == "json":
+        output = [asdict(item) for item in results]
+        print(json.dumps(output, indent=2, sort_keys=True))
+    else:
+        print_bench_results(results)
+
+
 def containers_parser(subparsers):
     parser = subparsers.add_parser("containers", aliases=["ps"], help="list all RamaLama containers")
     parser.add_argument(
@@ -508,6 +612,7 @@ def containers_parser(subparsers):
 
 
 def list_containers(args):
+
     containers = engine.containers(args)
     if len(containers) == 0:
         return
@@ -555,6 +660,7 @@ def human_readable_size(size):
 
 
 def _list_models_from_store(args):
+
     models = GlobalModelStore(args.store).list_models(engine=args.engine, show_container=args.container)
 
     ret = []
@@ -591,21 +697,22 @@ def _list_models(args):
     return _list_models_from_store(args)
 
 
-def info_cli(args):
-    info = {
+def info_cli(args: DefaultArgsType) -> None:
+    shortnames = get_shortnames()
+    info: dict[str, Any] = {
         "Accelerator": get_accel(),
         "Config": load_file_config(),
         "Engine": {
             "Name": args.engine,
         },
-        "Image": accel_image(CONFIG),
+        "Image": default_image(),
         "Inference": {
             "Default": args.runtime,
             "Engines": {spec: str(path) for spec, path in get_inference_spec_files().items()},
             "Schema": {schema: str(path) for schema, path in get_inference_schema_files().items()},
         },
-        "RagImage": rag_image(CONFIG),
-        "Selinux": CONFIG.selinux,
+        "RagImage": default_rag_image(),
+        "Selinux": get_config().selinux,
         "Shortnames": {
             "Files": shortnames.paths,
             "Names": shortnames.shortnames,
@@ -615,6 +722,8 @@ def info_cli(args):
         "Version": version(),
     }
     if args.engine and len(args.engine) > 0:
+        from ramalama import engine
+
         info["Engine"]["Info"] = engine.info(args)
 
     print(json.dumps(info, sort_keys=True, indent=4))
@@ -667,6 +776,7 @@ def help_cli(args):
 
 
 def pull_parser(subparsers):
+    config = get_config()
     parser = subparsers.add_parser("pull", help="pull AI Model from Model registry to local storage")
     parser.add_argument("--authfile", help="path of the authentication file")
     parser.add_argument(
@@ -677,7 +787,7 @@ def pull_parser(subparsers):
     )
     parser.add_argument(
         "--verify",
-        default=CONFIG.verify,
+        default=config.verify,
         action=CoerceToBool,
         help="verify the model after pull, disable to allow pulling of models with different endianness",
     )
@@ -686,11 +796,13 @@ def pull_parser(subparsers):
 
 
 def pull_cli(args):
+
     model = New(args.MODEL, args)
     model.pull(args)
 
 
 def convert_parser(subparsers):
+    config = get_config()
     parser = subparsers.add_parser(
         "convert",
         help="convert AI Model from local storage to OCI Image",
@@ -698,28 +810,28 @@ def convert_parser(subparsers):
     )
     parser.add_argument(
         "--carimage",
-        default=CONFIG.carimage,
+        default=config.carimage,
         help=argparse.SUPPRESS,
     )
     parser.add_argument(
         "--gguf",
         choices=get_args(GGUF_QUANTIZATION_MODES),
         nargs="?",
-        const=CONFIG.gguf_quantization_mode,  # Used if --gguf is provided without value
+        const=config.gguf_quantization_mode,  # Used if --gguf is provided without value
         default=None,  # Used if --gguf is not provided
-        help=f"GGUF quantization format. If specified without value, {CONFIG.gguf_quantization_mode} is used.",
+        help=f"GGUF quantization format. If specified without value, {config.gguf_quantization_mode} is used.",
     )
     add_network_argument(parser)
     parser.add_argument(
         "--rag-image",
-        default=rag_image(CONFIG),
+        default=default_rag_image(),
         help="Image to use for conversion to GGUF",
         action=OverrideDefaultAction,
         completer=local_images,
     )
     parser.add_argument(
         "--image",
-        default=accel_image(CONFIG),
+        default=default_image(),
         help="Image to use for quantization",
         action=OverrideDefaultAction,
         completer=local_images,
@@ -728,13 +840,13 @@ def convert_parser(subparsers):
         "--pull",
         dest="pull",
         type=str,
-        default=CONFIG.pull,
+        default=config.pull,
         choices=["always", "missing", "never", "newer"],
         help="pull image policy",
     )
     parser.add_argument(
         "--type",
-        default=CONFIG.convert_type,
+        default=config.convert_type,
         choices=["artifact", "car", "raw"],
         help="""\
 type of OCI Model Image to push.
@@ -753,6 +865,7 @@ def convert_cli(args):
         raise ValueError("convert command cannot be run with the --nocontainer option.")
 
     target = args.TARGET
+    shortnames = get_shortnames()
     tgt = shortnames.resolve(target)
 
     model = TransportFactory(tgt, args).create_oci()
@@ -762,6 +875,7 @@ def convert_cli(args):
 
 
 def push_parser(subparsers):
+    config = get_config()
     parser = subparsers.add_parser(
         "push",
         help="push AI Model from local storage to remote registry",
@@ -770,13 +884,13 @@ def push_parser(subparsers):
     parser.add_argument("--authfile", help="path of the authentication file")
     parser.add_argument(
         "--carimage",
-        default=CONFIG.carimage,
+        default=config.carimage,
         help=argparse.SUPPRESS,
     )
     add_network_argument(parser)
     parser.add_argument(
         "--type",
-        default=CONFIG.convert_type,
+        default=config.convert_type,
         choices=["artifact", "car", "raw"],
         help="""\
 type of OCI Model Image to push.
@@ -797,7 +911,9 @@ Model "raw" contains the model and a link file model.file to it stored at /.""",
 
 
 def _get_source_model(args, transport=None):
+    shortnames = get_shortnames()
     src = shortnames.resolve(args.SOURCE)
+
     smodel = New(src, args, transport=transport)
     if smodel.type == "OCI":
         if not args.TARGET:
@@ -809,6 +925,7 @@ def _get_source_model(args, transport=None):
 
 
 def push_cli(args):
+
     target = args.SOURCE
     transport = None
     if not args.TARGET:
@@ -816,7 +933,9 @@ def push_cli(args):
     source_model = _get_source_model(args, transport=transport)
 
     if args.TARGET:
+        shortnames = get_shortnames()
         target = shortnames.resolve(args.TARGET)
+
     target_model = New(target, args)
 
     try:
@@ -835,10 +954,11 @@ def push_cli(args):
 
 
 def runtime_options(parser, command):
+    config = get_config()
     if command in ["run", "serve"]:
         parser.add_argument(
             "--api",
-            default=CONFIG.api,
+            default=config.api,
             choices=["llama-stack", "none"],
             help="unified API layer for for Inference, RAG, Agents, Tools, Safety, Evals, and Telemetry.",
         )
@@ -848,7 +968,7 @@ def runtime_options(parser, command):
             "--cache-reuse",
             dest="cache_reuse",
             type=int,
-            default=CONFIG.cache_reuse,
+            default=config.cache_reuse,
             help="min chunk size to attempt reusing from the cache via KV shifting",
             completer=suppressCompleter,
         )
@@ -857,7 +977,7 @@ def runtime_options(parser, command):
             "--ctx-size",
             dest="context",
             type=int,
-            default=CONFIG.ctx_size,
+            default=config.ctx_size,
             help="size of the prompt context (0 = loaded from model)",
             completer=suppressCompleter,
         )
@@ -865,7 +985,7 @@ def runtime_options(parser, command):
             "--max-model-len",
             dest="context",
             type=int,
-            default=CONFIG.ctx_size,
+            default=config.ctx_size,
             help=argparse.SUPPRESS,
             completer=suppressCompleter,
         )
@@ -885,7 +1005,7 @@ def runtime_options(parser, command):
         dest="env",
         action='append',
         type=str,
-        default=CONFIG.env,
+        default=config.env,
         help="environment variables to add to the running container",
         completer=local_env,
     )
@@ -906,13 +1026,13 @@ def runtime_options(parser, command):
         )
         parser.add_argument(
             "--host",
-            default=CONFIG.host,
+            default=config.host,
             help="IP address to listen",
             completer=suppressCompleter,
         )
     parser.add_argument(
         "--image",
-        default=accel_image(CONFIG),
+        default=default_image(),
         help="OCI container image to run with the specified AI model",
         action=OverrideDefaultAction,
         completer=local_images,
@@ -920,7 +1040,7 @@ def runtime_options(parser, command):
     parser.add_argument(
         "--keep-groups",
         dest="podman_keep_groups",
-        default=CONFIG.keep_groups,
+        default=config.keep_groups,
         action="store_true",
         help="""pass `--group-add keep-groups` to podman.
 If GPU device on host is accessible to via group access, this option leaks the user groups into the container.""",
@@ -943,7 +1063,7 @@ If GPU device on host is accessible to via group access, this option leaks the u
             "--max-tokens",
             dest="max_tokens",
             type=int,
-            default=CONFIG.max_tokens,
+            default=config.max_tokens,
             help="maximum number of tokens to generate (0 = unlimited)",
             completer=suppressCompleter,
         )
@@ -952,13 +1072,13 @@ If GPU device on host is accessible to via group access, this option leaks the u
         "--ngl",
         dest="ngl",
         type=int,
-        default=CONFIG.ngl,
+        default=config.ngl,
         help="number of layers to offload to the gpu, if available",
         completer=suppressCompleter,
     )
     parser.add_argument(
         "--thinking",
-        default=CONFIG.thinking,
+        default=config.thinking,
         help="enable/disable thinking mode in reasoning models",
         action=CoerceToBool,
     )
@@ -972,7 +1092,7 @@ If GPU device on host is accessible to via group access, this option leaks the u
             "-p",
             "--port",
             type=parse_port_option,
-            default=CONFIG.port,
+            default=config.port,
             action=OverrideDefaultAction,
             help="port for AI Model server to listen on",
             completer=suppressCompleter,
@@ -984,7 +1104,7 @@ If GPU device on host is accessible to via group access, this option leaks the u
         "--pull",
         dest="pull",
         type=str,
-        default=CONFIG.pull,
+        default=config.pull,
         choices=["always", "missing", "never", "newer"],
         help='pull image policy',
     )
@@ -994,7 +1114,7 @@ If GPU device on host is accessible to via group access, this option leaks the u
         )
         parser.add_argument(
             "--rag-image",
-            default=rag_image(CONFIG),
+            default=default_rag_image(),
             help="OCI container image to run with the specified RAG data",
             action=OverrideDefaultAction,
             completer=local_images,
@@ -1011,13 +1131,14 @@ If GPU device on host is accessible to via group access, this option leaks the u
     parser.add_argument("--seed", help="override random seed", completer=suppressCompleter)
     parser.add_argument(
         "--selinux",
-        default=CONFIG.selinux,
+        default=config.selinux,
         action=CoerceToBool,
         help="Enable SELinux container separation",
     )
     parser.add_argument(
         "--temp",
-        default=CONFIG.temp,
+        type=float,
+        default=config.temp,
         help="temperature of the response from the AI model",
         completer=suppressCompleter,
     )
@@ -1054,14 +1175,15 @@ If GPU device on host is accessible to via group access, this option leaks the u
 
 
 def default_threads():
-    if CONFIG.threads < 0:
+    config = get_config()
+    if config.threads < 0:
         nproc = os.cpu_count()
         if nproc and nproc > 4:
             return int(nproc / 2)
 
         return 4
 
-    return CONFIG.threads
+    return config.threads
 
 
 def chat_run_options(parser):
@@ -1077,18 +1199,24 @@ def chat_run_options(parser):
     parser.add_argument(
         "--summarize-after",
         type=int,
-        default=CONFIG.summarize_after,
+        default=get_config().summarize_after,
         metavar="N",
         help="automatically summarize conversation history after N messages to prevent context growth (0=disabled)",
     )
 
 
+def _chat_cli(args):
+    from ramalama import chat as chat_module
+
+    return chat_module.chat(args)
+
+
 def chat_parser(subparsers):
     parser = subparsers.add_parser("chat", help="OpenAI chat with the specified RESTAPI URL")
     parser.add_argument(
         "--api-key",
         type=str,
-        default=CONFIG.api_key,
+        default=get_config().api_key,
         help="""OpenAI-compatible API key.
         Can also be set in ramalama.conf or via the RAMALAMA_API_KEY environment variable.""",
     )
@@ -1102,10 +1230,23 @@ def chat_parser(subparsers):
     parser.add_argument("--url", type=str, default="http://127.0.0.1:8080/v1", help="the url to send requests to")
     parser.add_argument("--model", "-m", type=str, completer=local_models, help="model for inferencing")
     parser.add_argument("--rag", type=str, help="a file or directory to use as context for the chat")
+    parser.add_argument(
+        "--max-tokens",
+        dest="max_tokens",
+        type=int,
+        default=get_config().max_tokens,
+        help="maximum number of tokens to generate (0 = unlimited)",
+    )
+    parser.add_argument(
+        "--temp",
+        type=float,
+        default=float(get_config().temp),
+        help="temperature of the response from the AI model",
+    )
     parser.add_argument(
         "ARGS", nargs="*", help="overrides the default prompt, and the output is returned without entering the chatbot"
     )
-    parser.set_defaults(func=chat.chat)
+    parser.set_defaults(func=_chat_cli)
 
 
 def _rag_args(args):
@@ -1146,10 +1287,10 @@ def run_parser(subparsers):
 
 
 def run_cli(args):
+
     try:
         # detect available port and update arguments
         args.port = compute_serving_port(args)
-
         model = New(args.MODEL, args)
         model.ensure_model_exists(args)
     except KeyError as e:
@@ -1161,14 +1302,20 @@ def run_cli(args):
         except Exception as exc:
             raise e from exc
 
+    is_api_transport = isinstance(model, APITransport)
+
+    if args.rag and is_api_transport:
+        raise ValueError("ramalama run --rag is not supported for hosted API transports.")
+
     if args.rag:
         if not args.container:
             raise ValueError("ramalama run --rag cannot be run with the --nocontainer option.")
         args = _rag_args(args)
-        model = RagTransport(model, assemble_command(args.model_args), args)
+
+        model = RagTransport(model, assemble_command_lazy(args.model_args), args)
         model.ensure_model_exists(args)
 
-    model.run(args, assemble_command(args))
+    model.run(args, assemble_command_lazy(args))
 
 
 def serve_parser(subparsers):
@@ -1179,6 +1326,7 @@ def serve_parser(subparsers):
 
 
 def serve_cli(args):
+
     if not args.container:
         args.detach = False
 
@@ -1208,14 +1356,18 @@ def serve_cli(args):
         except Exception:
             raise e
 
+    if isinstance(model, APITransport):
+        raise ValueError("ramalama serve is not supported for hosted API transports.")
+
     if args.rag:
         if not args.container:
             raise ValueError("ramalama serve --rag cannot be run with the --nocontainer option.")
         args = _rag_args(args)
-        model = RagTransport(model, assemble_command(args.model_args), args)
+
+        model = RagTransport(model, assemble_command_lazy(args.model_args), args)
         model.ensure_model_exists(args)
 
-    model.serve(args, assemble_command(args))
+    model.serve(args, assemble_command_lazy(args))
 
 
 def stop_parser(subparsers):
@@ -1229,6 +1381,8 @@ def stop_parser(subparsers):
 
 
 def stop_container(args):
+    from ramalama import engine
+
     if not args.all:
         engine.stop_container(args, args.NAME)
         return
@@ -1242,6 +1396,7 @@ def stop_container(args):
 
 
 def daemon_parser(subparsers) -> None:
+    config = get_config()
     parser: ArgumentParserWithDefaults = subparsers.add_parser("daemon", help="daemon operations")
     parser.set_defaults(func=lambda _: parser.print_help())
 
@@ -1250,7 +1405,7 @@ def daemon_parser(subparsers) -> None:
     start_parser = daemon_parsers.add_parser("start")
     start_parser.add_argument(
         "--image",
-        default=accel_image(CONFIG),
+        default=default_image(),
         help="OCI container image to run with the specified AI model",
         action=OverrideDefaultAction,
         completer=local_images,
@@ -1259,13 +1414,13 @@ def daemon_parser(subparsers) -> None:
         "--pull",
         dest="pull",
         type=str,
-        default=CONFIG.pull,
+        default=config.pull,
         choices=["always", "missing", "never", "newer"],
         help='pull image policy',
     )
     start_parser.add_argument(
         "--host",
-        default=CONFIG.host,
+        default=config.host,
         help="IP address to listen",
         completer=suppressCompleter,
     )
@@ -1273,7 +1428,7 @@ def daemon_parser(subparsers) -> None:
         "-p",
         "--port",
         type=parse_port_option,
-        default=CONFIG.port,
+        default=config.port,
         help="port for AI Model server to listen on",
         completer=suppressCompleter,
     )
@@ -1282,7 +1437,7 @@ def daemon_parser(subparsers) -> None:
     run_parser = daemon_parsers.add_parser("run")
     run_parser.add_argument(
         "--host",
-        default=CONFIG.host,
+        default=config.host,
         help="IP address to listen",
         completer=suppressCompleter,
     )
@@ -1290,7 +1445,7 @@ def daemon_parser(subparsers) -> None:
         "-p",
         "--port",
         type=parse_port_option,
-        default=CONFIG.port,
+        default=config.port,
         help="port for AI Model server to listen on",
         completer=suppressCompleter,
     )
@@ -1298,8 +1453,6 @@ def daemon_parser(subparsers) -> None:
 
 
 def daemon_start_cli(args):
-    from ramalama.common import exec_cmd
-
     daemon_cmd = []
     daemon_model_store_dir = args.store
     is_daemon_in_container = args.container and args.engine in get_args(SUPPORTED_ENGINES)
@@ -1330,7 +1483,7 @@ def daemon_start_cli(args):
         "--port",
         "8080" if is_daemon_in_container else args.port,
         "--host",
-        CONFIG.host if is_daemon_in_container else args.host,
+        get_config().host if is_daemon_in_container else args.host,
     ]
     exec_cmd(daemon_cmd)
 
@@ -1348,17 +1501,21 @@ def version_parser(subparsers):
 
 class AddPathOrUrl(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
+        if not isinstance(values, list):
+            raise ValueError("AddPathOrUrl can only be used with the settings `nargs='+'`")
+
         setattr(namespace, self.dest, [])
         namespace.urls = []
         for value in values:
             parsed = urlparse(value)
-            if parsed.scheme in ["file", ""] and parsed.netloc == "":
-                getattr(namespace, self.dest).append(parsed.path.rstrip("/"))
-            else:
+            if parsed.scheme in ["http", "https"]:
                 namespace.urls.append(value)
+            else:
+                getattr(namespace, self.dest).append(file_uri_to_path(value))
 
 
 def rag_parser(subparsers):
+    config = get_config()
     parser = subparsers.add_parser(
         "rag",
         help="generate and convert retrieval augmented generation (RAG) data from provided documents into an OCI Image",
@@ -1368,19 +1525,19 @@ def rag_parser(subparsers):
         dest="env",
         action='append',
         type=str,
-        default=CONFIG.env,
+        default=config.env,
         help="environment variables to add to the running RAG container",
         completer=local_env,
     )
     parser.add_argument(
         "--format",
-        default=CONFIG.rag_format,
+        default=config.rag_format,
         help="Output format for RAG Data",
         choices=["qdrant", "json", "markdown", "milvus"],
     )
     parser.add_argument(
         "--image",
-        default=rag_image(CONFIG),
+        default=default_rag_image(),
         help="Image to use for generating RAG data",
         action=OverrideDefaultAction,
         completer=local_images,
@@ -1388,7 +1545,7 @@ def rag_parser(subparsers):
     parser.add_argument(
         "--keep-groups",
         dest="podman_keep_groups",
-        default=CONFIG.keep_groups,
+        default=config.keep_groups,
         action="store_true",
         help="""pass `--group-add keep-groups` to podman.
 If GPU device on host is accessible to via group access, this option leaks the user groups into the container.""",
@@ -1398,24 +1555,22 @@ If GPU device on host is accessible to via group access, this option leaks the u
         "--pull",
         dest="pull",
         type=str,
-        default=CONFIG.pull,
+        default=config.pull,
         choices=["always", "missing", "never", "newer"],
         help='pull image policy',
     )
     parser.add_argument(
         "--selinux",
-        default=CONFIG.selinux,
+        default=config.selinux,
         action=CoerceToBool,
         help="Enable SELinux container separation",
     )
     parser.add_argument(
         "PATHS",
         nargs="+",
-        help=dedent(
-            """
+        help=dedent("""
         Files/URLs/Directory containing PDF, DOCX, PPTX, XLSX, HTML, AsciiDoc & Markdown
-        formatted files to be processed"""
-        ),
+        formatted files to be processed"""),
         action=AddPathOrUrl,
     )
     parser.add_argument(
@@ -1424,7 +1579,7 @@ If GPU device on host is accessible to via group access, this option leaks the u
     parser.add_argument(
         "--ocr",
         dest="ocr",
-        default=CONFIG.ocr,
+        default=config.ocr,
         action="store_true",
         help="Enable embedded image text extraction from PDF (Increases RAM Usage significantly)",
     )
@@ -1432,9 +1587,11 @@ If GPU device on host is accessible to via group access, this option leaks the u
 
 
 def rag_cli(args):
-    rag = Rag(args.DESTINATION)
-    args.inputdir = INPUT_DIR
-    rag.generate(args, assemble_command(args))
+    from ramalama import rag as rag_module
+
+    rag = rag_module.Rag(args.DESTINATION)
+    args.inputdir = rag_module.INPUT_DIR
+    rag.generate(args, assemble_command_lazy(args))
 
 
 def rm_parser(subparsers):
@@ -1456,6 +1613,8 @@ def _rm_oci_model(model, args) -> bool:
 
 def _rm_model(models, args):
     exceptions = []
+    shortnames = get_shortnames()
+
     for model in models:
         model = shortnames.resolve(model)
 
@@ -1516,9 +1675,10 @@ def perplexity_parser(subparsers):
 
 
 def perplexity_cli(args):
+
     model = New(args.MODEL, args)
     model.ensure_model_exists(args)
-    model.perplexity(args, assemble_command(args))
+    model.perplexity(args, assemble_command_lazy(args))
 
 
 def inspect_parser(subparsers):
@@ -1539,12 +1699,13 @@ def inspect_parser(subparsers):
 
 def inspect_cli(args):
     args.pull = "never"
+
     model = New(args.MODEL, args)
     print(model.inspect(args.all, args.get == "all", args.get, args.json, args.dryrun))
 
 
 def main() -> None:
-    def eprint(e, exit_code):
+    def eprint(e: Exception | str, exit_code: int):
         try:
             if args.debug:
                 logger.exception(e)
@@ -1575,8 +1736,8 @@ def main() -> None:
     except FileNotFoundError as e:
         eprint(e, errno.ENOENT)
     except HelpException:
-        parser.print_help()
-    except (IsADirectoryError, ConnectionError, IndexError, KeyError, ValueError, NoRefFileFound) as e:
+        parser.print_help()  # type: ignore[possibly-unbound]
+    except (ConnectionError, IndexError, KeyError, ValueError, NoRefFileFound) as e:
         eprint(e, errno.EINVAL)
     except NotImplementedError as e:
         eprint(e, errno.ENOSYS)
@@ -1593,12 +1754,16 @@ def main() -> None:
     except ParseError as e:
         eprint(f"Failed to parse model: {e}", errno.EINVAL)
     except SafetensorModelNotSupported:
-        eprint(
-            f"""Safetensor models are not supported. Please convert it to GGUF via:
-$ ramalama convert --gguf=<quantization> {args.model} <oci-name>
-$ ramalama run <oci-name>
-""",
-            errno.ENOTSUP,
+        message = (
+            "Safetensor models are not supported. Please convert it to GGUF via:\n"
+            f"$ ramalama convert --gguf=<quantization> {args.model} <oci-name>\n"  # type: ignore[possibly-unbound]
+            "$ ramalama run <oci-name>\n"
         )
+        eprint(message, errno.ENOTSUP)
     except NoGGUFModelFileFound:
-        eprint(f"No GGUF model file found for downloaded model '{args.model}'", errno.ENOENT)
+        eprint(f"No GGUF model file found for downloaded model '{args.model}'", errno.ENOENT)  # type: ignore
+    except Exception as e:
+        if isinstance(e, OSError) and hasattr(e, "winerror") and e.winerror == 206:
+            eprint("Path too long, please enable long path support in the Windows registry", errno.ENAMETOOLONG)
+        else:
+            raise
