@@ -1,3 +1,5 @@
+import copy
+import json
 import os
 import platform
 import random
@@ -6,9 +8,32 @@ import subprocess
 import sys
 import time
 from abc import ABC, abstractmethod
-from typing import Any, Dict, Optional
+from functools import cached_property
+from typing import TYPE_CHECKING, Any, Optional
 
-import ramalama.chat as chat
+from ramalama import chat
+from ramalama.common import ContainerEntryPoint
+from ramalama.compose import Compose
+from ramalama.config import get_config
+from ramalama.engine import Engine, dry_run, is_healthy, wait_for_healthy
+from ramalama.kube import Kube
+from ramalama.model_inspect.base_info import ModelInfoBase
+from ramalama.model_inspect.gguf_info import GGUFModelInfo
+from ramalama.model_inspect.gguf_parser import GGUFInfoParser
+from ramalama.model_inspect.safetensor_info import SafetensorModelInfo
+from ramalama.model_inspect.safetensor_parser import SafetensorInfoParser
+from ramalama.model_store.global_store import GlobalModelStore
+from ramalama.model_store.store import ModelStore
+from ramalama.quadlet import Quadlet
+
+if TYPE_CHECKING:
+    from ramalama.chat import ChatOperationalArgs
+
+from datetime import datetime, timezone
+
+from ramalama.benchmarks.manager import BenchmarksManager
+from ramalama.benchmarks.schemas import BenchmarkRecord, BenchmarkRecordV1, get_benchmark_record
+from ramalama.benchmarks.utilities import parse_json, print_bench_results
 from ramalama.common import (
     MNT_DIR,
     MNT_FILE_DRAFT,
@@ -18,22 +43,11 @@ from ramalama.common import (
     is_split_file_model,
     perror,
     populate_volume_from_image,
+    run_cmd,
     set_accel_env_vars,
 )
-from ramalama.compose import Compose
-from ramalama.config import CONFIG, DEFAULT_PORT_RANGE
-from ramalama.engine import Engine, dry_run, is_healthy, wait_for_healthy
-from ramalama.kube import Kube
 from ramalama.logger import logger
-from ramalama.model_inspect.base_info import ModelInfoBase
-from ramalama.model_inspect.gguf_info import GGUFModelInfo
-from ramalama.model_inspect.gguf_parser import GGUFInfoParser
-from ramalama.model_inspect.safetensor_info import SafetensorModelInfo
-from ramalama.model_inspect.safetensor_parser import SafetensorInfoParser
-from ramalama.model_store.global_store import GlobalModelStore
-from ramalama.model_store.store import ModelStore
 from ramalama.path_utils import get_container_mount_path
-from ramalama.quadlet import Quadlet
 
 MODEL_TYPES = ["file", "https", "http", "oci", "huggingface", "hf", "modelscope", "ms", "ollama", "rlcr"]
 
@@ -100,7 +114,7 @@ class TransportBase(ABC):
         raise self.__not_implemented_error("push")
 
     @abstractmethod
-    def remove(self, args):
+    def remove(self, args) -> bool:
         raise self.__not_implemented_error("rm")
 
     @abstractmethod
@@ -108,7 +122,7 @@ class TransportBase(ABC):
         raise self.__not_implemented_error("bench")
 
     @abstractmethod
-    def run(self, args, server_cmd: list[str]):
+    def run(self, args, cmd: list[str]):
         raise self.__not_implemented_error("run")
 
     @abstractmethod
@@ -148,11 +162,15 @@ class Transport(TransportBase):
         self._model_type = type(self).__name__.lower()
 
         self._model_store_path: str = model_store_path
-        self._model_store: Optional[ModelStore] = None
+        self._model_store: Optional["ModelStore"] = None
 
-        self.default_image = accel_image(CONFIG)
+        self.default_image = accel_image(get_config())
         self.draft_model: Transport | None = None
 
+    @cached_property
+    def artifact(self) -> bool:
+        return self.is_artifact()
+
     def extract_model_identifiers(self):
         model_name = self.model
         model_tag = "latest"
@@ -160,7 +178,7 @@ class Transport(TransportBase):
 
         # extract model tag from name if exists
         if ":" in model_name:
-            model_name, model_tag = model_name.split(":", 1)
+            model_name, model_tag = model_name.rsplit(":", 1)
 
         # extract model organization from name if exists and update name
         split = model_name.rsplit("/", 1)
@@ -173,6 +191,10 @@ class Transport(TransportBase):
     def model_name(self) -> str:
         return self._model_name
 
+    @property
+    def model_alias(self):
+        return f"{self.model_organization}/{self.model_name}"
+
     @property
     def model_tag(self) -> str:
         return self._model_tag
@@ -186,12 +208,61 @@ class Transport(TransportBase):
         return self._model_type
 
     @property
-    def model_store(self) -> ModelStore:
+    def model_store(self) -> "ModelStore":
         if self._model_store is None:
             name, _, orga = self.extract_model_identifiers()
             self._model_store = ModelStore(GlobalModelStore(self._model_store_path), name, self.model_type, orga)
         return self._model_store
 
+    def _get_all_model_part_paths(
+        self, use_container: bool, should_generate: bool, dry_run: bool
+    ) -> list[tuple[str, str]]:
+        """
+        Returns a list of (src_path, dest_path) tuples for all parts of a model.
+        For single-file models, returns a list with one tuple.
+        For multi-part models, returns a tuple for each part.
+        """
+        if dry_run:
+            return [("/path/to/model", f"{MNT_DIR}/model.file")]
+
+        if self.model_type == 'oci':
+            # OCI models don't use this path for multi-part handling
+            entry_path_src = self._get_entry_model_path(False, False, False)
+            entry_path_dest = self._get_entry_model_path(True, True, False)
+            return [(entry_path_src, entry_path_dest)]
+
+        ref_file = self.model_store.get_ref_file(self.model_tag)
+        if ref_file is None:
+            raise NoRefFileFound(self.model)
+
+        gguf_files = ref_file.model_files
+        safetensor_files = ref_file.safetensor_model_files
+        if safetensor_files:
+            # Safetensor models use directory mounts, not individual files
+            src_path = self.model_store.get_snapshot_directory_from_tag(self.model_tag)
+            if use_container or should_generate:
+                dest_path = MNT_DIR
+            else:
+                dest_path = src_path
+            return [(src_path, dest_path)]
+        elif not gguf_files:
+            raise NoGGUFModelFileFound()
+
+        model_parts = []
+        for model_file in gguf_files:
+            if use_container or should_generate:
+                dest_path = f"{MNT_DIR}/{model_file.name}"
+            else:
+                dest_path = self.model_store.get_blob_file_path(model_file.hash)
+            src_path = self.model_store.get_blob_file_path(model_file.hash)
+            model_parts.append((src_path, dest_path))
+
+        # Sort multi-part models by filename to ensure correct order
+        if len(model_parts) > 1 and any("-00001-of-" in name for _, name in model_parts):
+            model_parts.sort(key=lambda x: x[1])
+
+        return model_parts
+
     def _get_entry_model_path(self, use_container: bool, should_generate: bool, dry_run: bool) -> str:
         """
         Returns the path to the model blob on the host if use_container and should_generate are both False.
@@ -202,6 +273,13 @@ class Transport(TransportBase):
 
         if self.model_type == 'oci':
             if use_container or should_generate:
+                if getattr(self, "artifact", False):
+                    artifact_name_method = getattr(self, "artifact_name", None)
+                    if artifact_name_method:
+                        try:
+                            return f"{MNT_DIR}/{artifact_name_method()}"
+                        except subprocess.CalledProcessError:
+                            pass
                 return f"{MNT_DIR}/model.file"
             else:
                 return f"oci://{self.model}"
@@ -291,10 +369,13 @@ class Transport(TransportBase):
             return f"{MNT_DIR}/{chat_template_file.name}"
         return self.model_store.get_blob_file_path(chat_template_file.hash)
 
-    def remove(self, args):
+    def remove(self, args) -> bool:
         _, tag, _ = self.extract_model_identifiers()
-        if not self.model_store.remove_snapshot(tag) and not args.ignore:
+        if self.model_store.remove_snapshot(tag):
+            return True
+        if not args.ignore:
             raise KeyError(f"Model '{self.model}' not found")
+        return False
 
     def get_container_name(self, args):
         if getattr(args, "name", None):
@@ -302,7 +383,7 @@ class Transport(TransportBase):
 
         return genname()
 
-    def new_engine(self, args):
+    def new_engine(self, args) -> Engine:
         return Engine(args)
 
     def base(self, args, name):
@@ -332,6 +413,10 @@ class Transport(TransportBase):
         if not args.container:
             return False
 
+        if len(cmd_args) > 0 and isinstance(cmd_args[0], ContainerEntryPoint):
+            # Ignore entrypoint
+            cmd_args = cmd_args[1:]
+
         self.setup_container(args)
         self.setup_mounts(args)
 
@@ -347,9 +432,10 @@ class Transport(TransportBase):
     def setup_mounts(self, args):
         if args.dryrun:
             return
+
         if self.model_type == 'oci':
             if self.engine.use_podman:
-                mount_cmd = f"--mount=type=image,src={self.model},destination={MNT_DIR},subpath=/models,rw=false"
+                mount_cmd = self.mount_cmd()
             elif self.engine.use_docker:
                 output_filename = self._get_entry_model_path(args.container, True, args.dryrun)
                 volume = populate_volume_from_image(self, args, os.path.basename(output_filename))
@@ -384,60 +470,135 @@ class Transport(TransportBase):
 
     def bench(self, args, cmd: list[str]):
         set_accel_env_vars()
-        self.execute_command(cmd, args)
 
-    def run(self, args, server_cmd: list[str]):
+        output_format = getattr(args, "format", "table")
+
+        if args.dryrun:
+            if args.container:
+                self.engine.dryrun()
+            else:
+                dry_run(cmd)
+
+            return
+        elif args.container:
+            self.setup_container(args)
+            self.setup_mounts(args)
+            self.engine.add([args.image] + cmd)
+            result = self.engine.run_process()
+        else:
+            result = run_cmd(cmd, encoding="utf-8")
+
+        try:
+            bench_results = parse_json(result.stdout)
+        except (json.JSONDecodeError, ValueError):
+            message = f"Could not parse benchmark output. Expected JSON but got:\n{result.stdout}"
+            raise ValueError(message)
+
+        base_payload: dict = {
+            "created_at": datetime.now(timezone.utc).isoformat(),
+            "configuration": {
+                "container_image": args.image,
+                "container_runtime": args.engine,
+                "inference_engine": args.runtime,
+                "runtime_args": cmd,
+            },
+        }
+        results: list[BenchmarkRecord] = list()
+        for bench_result in bench_results:
+            result_record: BenchmarkRecordV1 = get_benchmark_record({"result": bench_result, **base_payload}, "v1")
+            results.append(result_record)
+
+        if output_format == "json":
+            print(result.stdout)
+        else:
+            print_bench_results(results)
+
+        config = get_config()
+        if not config.benchmarks.disable:
+            bench_manager = BenchmarksManager(config.benchmarks.storage_folder)
+            bench_manager.save(results)
+
+    def run(self, args, cmd: list[str]):
         # The Run command will first launch a daemonized service
         # and run chat to communicate with it.
 
-        args.noout = not args.debug
+        if len(cmd) > 0 and isinstance(cmd[0], ContainerEntryPoint):
+            # Ignore entrypoint
+            cmd = cmd[1:]
 
-        pid = self._fork_and_serve(args, server_cmd)
-        if pid:
-            return self._connect_and_chat(args, pid)
+        process = self.serve_nonblocking(args, cmd)
+        if process:
+            return self._connect_and_chat(args, process)
 
-    def _fork_and_serve(self, args, cmd: list[str]):
+    def serve_nonblocking(self, args, cmd: list[str]) -> subprocess.Popen | None:
         if args.container:
             args.name = self.get_container_name(args)
-        pid = os.fork()
-        if pid == 0:
-            # Child process - start the server
-            self._start_server(args, cmd)
-        return pid
 
-    def _start_server(self, args, cmd: list[str]):
-        """Start the server in the child process."""
-        args.host = CONFIG.host
-        args.generate = ""
+        args.host = get_config().host
         args.detach = True
-        self.serve(args, cmd)
 
-    def _connect_and_chat(self, args, server_pid):
+        set_accel_env_vars()
+
+        if args.container:
+            # For container mode, set up the container and start it with subprocess
+            self.setup_container(args)
+            self.setup_mounts(args)
+            # Make sure Image precedes cmd_args
+            self.engine.add([args.image] + cmd)
+
+            if args.dryrun:
+                self.engine.dryrun()
+                return None
+
+            # Start the container using subprocess.Popen
+            process = subprocess.Popen(
+                self.engine.exec_args,
+            )
+            return process
+
+        # Non-container mode: run the command directly with subprocess
+        if args.dryrun:
+            dry_run(cmd)
+            return None
+
+        process = subprocess.Popen(
+            cmd,
+        )
+        return process
+
+    def _connect_and_chat(self, args, server_process):
         """Connect to the server and start chat in the parent process."""
         args.url = f"http://127.0.0.1:{args.port}/v1"
         if getattr(args, "runtime", None) == "mlx":
             args.prefix = "🍏 > "
-        args.pid2kill = ""
+
+        # Model name in the chat request must match RamalamaModelContext.alias()
+        chat_args = copy.deepcopy(args)
+        chat_args.model = f"{self.model_organization}/{self.model_name}"
 
         if args.container:
-            return self._handle_container_chat(args, server_pid)
+            return self._handle_container_chat(chat_args, server_process)
         else:
-            args.pid2kill = server_pid
-            if getattr(args, "runtime", None) == "mlx":
-                return self._handle_mlx_chat(args)
-            chat.chat(args)
+            # Store the Popen object for monitoring
+            chat_args.server_process = server_process
+
+            if getattr(chat_args, "runtime", None) == "mlx":
+                return self._handle_mlx_chat(chat_args)
+            chat.chat(chat_args)
             return 0
 
-    def chat_operational_args(self, args):
+    def chat_operational_args(self, args) -> "ChatOperationalArgs | None":
         return None
 
     def wait_for_healthy(self, args):
         wait_for_healthy(args, is_healthy)
 
-    def _handle_container_chat(self, args, server_pid):
+    def _handle_container_chat(self, args, server_process):
         """Handle chat for container-based execution."""
-        _, status = os.waitpid(server_pid, 0)
-        if status != 0:
+
+        # Wait for the server process to complete (blocking)
+        exit_code = server_process.wait()
+        if exit_code != 0:
             raise ValueError(f"Failed to serve model {self.model_name}, for ramalama run command")
 
         if not args.dryrun:
@@ -480,7 +641,7 @@ class Transport(TransportBase):
             except Exception as e:
                 if i >= max_retries - 1:
                     perror(f"Error: Failed to connect to MLX server after {max_retries} attempts: {e}")
-                    self._cleanup_server_process(args.pid2kill)
+                    self._cleanup_server_process(args.server_process)
                     raise e
                 logger.debug(f"Connection attempt failed, retrying... (attempt {i + 1}/{max_retries}): {e}")
                 time.sleep(3)
@@ -498,32 +659,16 @@ class Transport(TransportBase):
         except (socket.error, ValueError):
             return False
 
-    def _cleanup_server_process(self, pid):
+    def _cleanup_server_process(self, process):
         """Clean up the server process."""
-        if not pid:
+        if not process:
             return
 
-        import signal
-
+        process.terminate()
         try:
-            # Try graceful termination first
-            os.kill(pid, signal.SIGTERM)
-            time.sleep(1)  # Give it time to terminate gracefully
-
-            # Force kill if still running (SIGKILL is Unix-only)
-            if hasattr(signal, 'SIGKILL'):
-                try:
-                    os.kill(pid, signal.SIGKILL)
-                except ProcessLookupError:
-                    pass
-            else:
-                # On Windows, send SIGTERM again as fallback
-                try:
-                    os.kill(pid, signal.SIGTERM)
-                except ProcessLookupError:
-                    pass
-        except ProcessLookupError:
-            pass
+            process.wait(timeout=5)
+        except subprocess.TimeoutExpired:
+            process.kill()
 
     def perplexity(self, args, cmd: list[str]):
         set_accel_env_vars()
@@ -577,6 +722,9 @@ class Transport(TransportBase):
         chat_template_dest_path = self._get_chat_template_path(True, True, args.dryrun)
         mmproj_dest_path = self._get_mmproj_path(True, True, args.dryrun)
 
+        # Get all model parts (for multi-part models)
+        model_parts = self._get_all_model_part_paths(False, True, args.dryrun)
+
         if args.generate.gen_type == "quadlet":
             self.quadlet(
                 (model_src_path, model_dest_path),
@@ -585,10 +733,11 @@ class Transport(TransportBase):
                 args,
                 exec_args,
                 args.generate.output_dir,
+                model_parts,
             )
         elif args.generate.gen_type == "kube":
             self.kube(
-                (model_src_path, model_dest_path),
+                (model_src_path.removeprefix("oci://"), model_dest_path),
                 (chat_template_src_path, chat_template_dest_path),
                 (mmproj_src_path, mmproj_dest_path),
                 args,
@@ -603,6 +752,7 @@ class Transport(TransportBase):
                 args,
                 exec_args,
                 args.generate.output_dir,
+                model_parts,
             )
         elif args.generate.gen_type == "compose":
             self.compose(
@@ -636,30 +786,41 @@ class Transport(TransportBase):
             self.generate_container_config(args, cmd)
             return
 
-        self.execute_command(cmd, args)
+        try:
+            self.execute_command(cmd, args)
+        except Exception as e:
+            self._cleanup_server_process(args.server_process)
+            raise e
 
-    def quadlet(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir):
-        quadlet = Quadlet(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args)
+    def quadlet(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir, model_parts=None):
+        quadlet = Quadlet(
+            self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.artifact, model_parts
+        )
         for generated_file in quadlet.generate():
             generated_file.write(output_dir)
 
-    def quadlet_kube(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir):
-        kube = Kube(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args)
+    def quadlet_kube(
+        self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir, model_parts=None
+    ):
+        kube = Kube(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.artifact)
         kube.generate().write(output_dir)
 
-        quadlet = Quadlet(kube.name, model_paths, chat_template_paths, mmproj_paths, args, exec_args)
+        quadlet = Quadlet(
+            kube.name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.artifact, model_parts
+        )
         quadlet.kube().write(output_dir)
 
     def kube(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir):
-        kube = Kube(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args)
+        kube = Kube(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.artifact)
         kube.generate().write(output_dir)
 
     def compose(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir):
         compose = Compose(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args)
         compose.generate().write(output_dir)
 
-    def inspect_metadata(self) -> Dict[str, Any]:
+    def inspect_metadata(self) -> dict[str, Any]:
         model_path = self._get_entry_model_path(False, False, False)
+
         if GGUFInfoParser.is_model_gguf(model_path):
             return GGUFInfoParser.parse_metadata(model_path).data
         return {}
@@ -671,45 +832,49 @@ class Transport(TransportBase):
         get_field: str = "",
         as_json: bool = False,
         dryrun: bool = False,
-    ) -> None:
+    ) -> Any:
         model_name = self.filename
         model_registry = self.type.lower()
         model_path = self._get_inspect_model_path(dryrun)
-
         if GGUFInfoParser.is_model_gguf(model_path):
             if not show_all_metadata and get_field == "":
                 gguf_info: GGUFModelInfo = GGUFInfoParser.parse(model_name, model_registry, model_path)
-                print(gguf_info.serialize(json=as_json, all=show_all))
-                return
+                return gguf_info.serialize(json=as_json, all=show_all)
 
             metadata = GGUFInfoParser.parse_metadata(model_path)
             if show_all_metadata:
-                print(metadata.serialize(json=as_json))
-                return
+                return metadata.serialize(json=as_json)
             elif get_field != "":  # If a specific field is requested, print only that field
                 field_value = metadata.get(get_field)
                 if field_value is None:
                     raise KeyError(f"Field '{get_field}' not found in GGUF model metadata")
-                print(field_value)
-                return
+                return field_value
 
         if SafetensorInfoParser.is_model_safetensor(model_name):
             safetensor_info: SafetensorModelInfo = SafetensorInfoParser.parse(model_name, model_registry, model_path)
-            print(safetensor_info.serialize(json=as_json, all=show_all))
-            return
+            return safetensor_info.serialize(json=as_json, all=show_all)
 
-        print(ModelInfoBase(model_name, model_registry, model_path).serialize(json=as_json))
+        return ModelInfoBase(model_name, model_registry, model_path).serialize(json=as_json)
 
-    def print_pull_message(self, model_name):
+    def print_pull_message(self, model_name) -> None:
         model_name = trim_model_name(model_name)
         # Write messages to stderr
         perror(f"Downloading {model_name} ...")
         perror(f"Trying to pull {model_name} ...")
 
+    def is_artifact(self) -> bool:
+        return False
+
 
 def compute_ports(exclude: list[str] | None = None) -> list[int]:
-    excluded = exclude and set(map(int, exclude)) or set()
-    ports = list(sorted(set(range(DEFAULT_PORT_RANGE[0], DEFAULT_PORT_RANGE[1] + 1)) - excluded))
+    excluded = set() if exclude is None else set(map(int, exclude))
+
+    port_range = get_config().default_port_range
+    ports = [p for p in range(port_range[0], port_range[1] + 1) if p not in excluded]
+
+    if not ports:
+        raise ValueError("All ports in the default port range were exhausted by the exclusion list.")
+
     first_port = ports.pop(0)
     random.shuffle(ports)
     # try always the first port before the randomized others
